# ---------------------------------------------------------------
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#
# This file has been modified from a file in the Score SDE library
# which was released under the Apache License.
#
# Source:
# https://github.com/yang-song/score_sde_pytorch/blob/main/models/layerspp.py
#
# The license for the original version of this file can be
# found in this directory (LICENSE_Apache). The modifications
# to this file are subject to the same Apache License.
# ---------------------------------------------------------------

# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file
''' Codes adapted from https://github.com/yang-song/score_sde_pytorch/blob/main/models/ncsnpp.py
'''

from . import utils, layers, layerspp, dense_layer
import torch.nn as nn
import functools
import torch
import numpy as np


ResnetBlockDDPM = layerspp.ResnetBlockDDPMpp_Adagn
ResnetBlockBigGAN = layerspp.ResnetBlockBigGANpp_Adagn
ResnetBlockBigGAN_one = layerspp.ResnetBlockBigGANpp_Adagn_one
Combine = layerspp.Combine
conv3x3 = layerspp.conv3x3
conv1x1 = layerspp.conv1x1
get_act = layers.get_act
default_initializer = layers.default_init
dense = dense_layer.dense

class PixelNorm(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return input / torch.sqrt(torch.mean(input ** 2, dim=1, keepdim=True) + 1e-8)


@utils.register_model(name='ncsnpp')
class NCSNpp(nn.Module):
  """NCSN++ model"""

  def __init__(self, config):
    super().__init__()
    self.config = config
    self.not_use_tanh = config.not_use_tanh
    self.act = act = nn.SiLU()
    self.z_emb_dim = z_emb_dim = config.z_emb_dim
    
    self.nf = nf = config.num_channels_dae
    ch_mult = config.ch_mult
    self.num_res_blocks = num_res_blocks = config.num_res_blocks
    self.attn_resolutions = attn_resolutions = config.attn_resolutions
    dropout = config.dropout
    resamp_with_conv = config.resamp_with_conv
    self.num_resolutions = num_resolutions = len(ch_mult)
    self.all_resolutions = all_resolutions = [config.image_size // (2 ** i) for i in range(num_resolutions)]

    self.conditional = conditional = config.conditional  # noise-conditional
    fir = config.fir
    fir_kernel = config.fir_kernel
    self.skip_rescale = skip_rescale = config.skip_rescale
    self.resblock_type = resblock_type = config.resblock_type.lower()
    self.progressive = progressive = config.progressive.lower()
    self.progressive_input = progressive_input = config.progressive_input.lower()
    self.embedding_type = embedding_type = config.embedding_type.lower()
    self.num_downsampling = config.num_downsampling
    init_scale = 0.
    assert progressive in ['none', 'output_skip', 'residual']
    assert progressive_input in ['none', 'input_skip', 'residual']
    assert embedding_type in ['fourier', 'positional']
    combine_method = config.progressive_combine.lower()
    combiner = functools.partial(Combine, method=combine_method)

    modules = []
    modules_emb = []
    # timestep/noise_level embedding; only for continuous training
    if embedding_type == 'fourier':
      # Gaussian Fourier features embeddings.
      #assert config.training.continuous, "Fourier features are only used for continuous training."

      modules.append(layerspp.GaussianFourierProjection(
        embedding_size=nf, scale=config.fourier_scale
      ))
      embed_dim = 2 * nf

    elif embedding_type == 'positional':
      embed_dim =  nf

    else:
      raise ValueError(f'embedding type {embedding_type} unknown.')

    if conditional:
      modules_emb.append(nn.Linear(embed_dim, nf * 4))
      modules_emb[-1].weight.data = default_initializer()(modules_emb[-1].weight.shape)
      nn.init.zeros_(modules_emb[-1].bias)
      modules_emb.append(nn.Linear(nf * 4, nf * 4))
      modules_emb[-1].weight.data = default_initializer()(modules_emb[-1].weight.shape)
      nn.init.zeros_(modules_emb[-1].bias)

    AttnBlock = functools.partial(layerspp.AttnBlockpp,
                                  init_scale=init_scale,
                                  skip_rescale=skip_rescale)

    Upsample = functools.partial(layerspp.Upsample,
                                 with_conv=resamp_with_conv, fir=fir, fir_kernel=fir_kernel)

    if progressive == 'output_skip':
      self.pyramid_upsample = layerspp.Upsample(fir=fir, fir_kernel=fir_kernel, with_conv=False)
    elif progressive == 'residual':
      pyramid_upsample = functools.partial(layerspp.Upsample,
                                           fir=fir, fir_kernel=fir_kernel, with_conv=True)

    Downsample = functools.partial(layerspp.Downsample,
                                   with_conv=resamp_with_conv, fir=fir, fir_kernel=fir_kernel)

    if progressive_input == 'input_skip':
      self.pyramid_downsample = layerspp.Downsample(fir=fir, fir_kernel=fir_kernel, with_conv=False)
    elif progressive_input == 'residual':
      pyramid_downsample = functools.partial(layerspp.Downsample,
                                             fir=fir, fir_kernel=fir_kernel, with_conv=True)

    if resblock_type == 'ddpm':
      ResnetBlock = functools.partial(ResnetBlockDDPM,
                                      act=act,
                                      dropout=dropout,
                                      init_scale=init_scale,
                                      skip_rescale=skip_rescale,
                                      temb_dim=nf * 4,
                                      zemb_dim = z_emb_dim)

    elif resblock_type == 'biggan':
      ResnetBlock = functools.partial(ResnetBlockBigGAN,
                                      act=act,
                                      dropout=dropout,
                                      fir=fir,
                                      fir_kernel=fir_kernel,
                                      init_scale=init_scale,
                                      skip_rescale=skip_rescale,
                                      temb_dim=nf * 4,
                                      zemb_dim = z_emb_dim)
    elif resblock_type == 'biggan_oneadagn':
      ResnetBlock = functools.partial(ResnetBlockBigGAN_one,
                                      act=act,
                                      dropout=dropout,
                                      fir=fir,
                                      fir_kernel=fir_kernel,
                                      init_scale=init_scale,
                                      skip_rescale=skip_rescale,
                                      temb_dim=nf * 4,
                                      zemb_dim = z_emb_dim)

    else:
      raise ValueError(f'resblock type {resblock_type} unrecognized.')

    # Downsampling block

    channels = config.num_channels
    if progressive_input != 'none':
      input_pyramid_ch = channels
    init_modules =[]
    init_modules.append(conv3x3(channels, nf))
    # modules.append(conv3x3(channels, nf))
    hs_c = [nf]

    in_ch = nf
    for i_level in range(self.num_downsampling):
      mult = 2**(i_level+1)
      out_ch = nf * mult
      modules.append(ResnetBlock(in_ch=in_ch, out_ch=out_ch,down=True))
      in_ch = out_ch
    for i_block in range(num_res_blocks):
      modules.append(ResnetBlock(in_ch=in_ch, out_ch=in_ch))
    for i_level in range(self.num_downsampling):
      # mult = 2**(i_level+1)
      out_ch = in_ch // 2
      modules.append(ResnetBlock(in_ch=in_ch, out_ch=out_ch,up=True))
      in_ch = out_ch
      # out_ch = nf*mult
      # Residual blocks for this resolution
#       if i_level is num_resolutions-1:
#         for i_block in range(num_res_blocks):
#           out_ch = nf * ch_mult[i_level]
#           modules.append(ResnetBlock(in_ch=in_ch, out_ch=out_ch,down=True))
#           in_ch = out_ch
#       else:
#         out_ch = nf * ch_mult[i_level]
#         modules.append(ResnetBlock(in_ch=in_ch, out_ch=out_ch))
#         in_ch = out_ch
#         if all_resolutions[i_level] in attn_resolutions:
#           modules.append(AttnBlock(channels=in_ch))
#         hs_c.append(in_ch)

#       if i_level != num_resolutions - 1:
#         if resblock_type == 'ddpm':
#           modules.append(Downsample(in_ch=in_ch))
#         else:
#           modules.append(ResnetBlock(down=True, in_ch=in_ch))

#         if progressive_input == 'input_skip':
#           modules.append(combiner(dim1=input_pyramid_ch, dim2=in_ch))
#           if combine_method == 'cat':
#             in_ch *= 2

#         elif progressive_input == 'residual':
#           modules.append(pyramid_downsample(in_ch=input_pyramid_ch, out_ch=in_ch))
#           input_pyramid_ch = in_ch

#         hs_c.append(in_ch)

#     in_ch = hs_c[-1]
#     modules.append(ResnetBlock(in_ch=in_ch))
#     # modules.append(AttnBlock(channels=in_ch))
#     modules.append(ResnetBlock(in_ch=in_ch))
#     # print(len(hs_c))
#     pyramid_ch = 0
#     # Upsampling block
#     for i_level in reversed(range(num_resolutions)):
#       # print(num_resolutions)
#       if i_level is num_resolutions-1:
#         for i_block in range(num_res_blocks):
#           out_ch = nf * ch_mult[i_level]
#           modules.append(ResnetBlock(in_ch=in_ch + hs_c.pop(),
#                                    out_ch=out_ch))
#           in_ch = out_ch
#       else:
#         out_ch = nf * ch_mult[i_level]
#         modules.append(ResnetBlock(in_ch=in_ch + hs_c.pop(),
#                                    out_ch=out_ch))
#         in_ch = out_ch
#       if all_resolutions[i_level] in attn_resolutions:
#         modules.append(AttnBlock(channels=in_ch))

#       if progressive != 'none':
#         if i_level == num_resolutions - 1:
#           if progressive == 'output_skip':
#             modules.append(nn.GroupNorm(num_groups=min(in_ch // 4, 32),
#                                         num_channels=in_ch, eps=1e-6))
#             modules.append(conv3x3(in_ch, channels, init_scale=init_scale))
#             pyramid_ch = channels
#           elif progressive == 'residual':
#             modules.append(nn.GroupNorm(num_groups=min(in_ch // 4, 32),
#                                         num_channels=in_ch, eps=1e-6))
#             modules.append(conv3x3(in_ch, in_ch, bias=True))
#             pyramid_ch = in_ch
#           else:
#             raise ValueError(f'{progressive} is not a valid name.')
#         else:
#           if progressive == 'output_skip':
#             modules.append(nn.GroupNorm(num_groups=min(in_ch // 4, 32),
#                                         num_channels=in_ch, eps=1e-6))
#             modules.append(conv3x3(in_ch, channels, bias=True, init_scale=init_scale))
#             pyramid_ch = channels
#           elif progressive == 'residual':
#             modules.append(pyramid_upsample(in_ch=pyramid_ch, out_ch=in_ch))
#             pyramid_ch = in_ch
#           else:
#             raise ValueError(f'{progressive} is not a valid name')

#       if i_level != 0:
#         if resblock_type == 'ddpm':
#           modules.append(Upsample(in_ch=in_ch))
#         else:
#           modules.append(ResnetBlock(in_ch=in_ch, up=True))

#     assert not hs_c
    fin_modules =[]
    if progressive != 'output_skip':
      fin_modules.append(nn.GroupNorm(num_groups=min(in_ch // 4, 32),
                                  num_channels=in_ch, eps=1e-6))
      fin_modules.append(conv3x3(in_ch, channels, init_scale=init_scale))
    self.embed_modules = nn.ModuleList(modules_emb)
    self.all_modules = nn.ModuleList(modules)
    self.fin_modules = nn.ModuleList(fin_modules)
    self.init_modules = nn.ModuleList(init_modules)
    mapping_layers = [PixelNorm(),
                      dense(config.nz, z_emb_dim),
                      self.act,]
    for _ in range(config.n_mlp):
        mapping_layers.append(dense(z_emb_dim, z_emb_dim))
        mapping_layers.append(self.act)
    self.z_transform = nn.Sequential(*mapping_layers)
    

  def forward(self, x, time_cond, z):
    # timestep/noise_level embedding; only for continuous training
    zemb = self.z_transform(z)
    # modules = self.all_modules
    m_idx = 0
    if self.embedding_type == 'fourier':
      # Gaussian Fourier features embeddings.
      used_sigmas = time_cond
      temb = modules[m_idx](torch.log(used_sigmas))
      m_idx += 1

    elif self.embedding_type == 'positional':
      # Sinusoidal positional embeddings.
      timesteps = time_cond
      # print(time_cond.shape)
      temb = layers.get_timestep_embedding(timesteps, self.nf)
      # print(temb.shape)
    else:
      raise ValueError(f'embedding type {self.embedding_type} unknown.')

    if self.conditional:
      for mod in self.embed_modules:
          # print(temb.shape)
        temb = mod(temb)
        # print(self.act(temb).shape)
          # m_idx += 1
        temb = self.act(temb)
        # temb = self.act(temb))
          # m_idx += 1
    else:
      temb = None

    if not self.config.centered:
      # If input data is in [0, 1]
      x = 2 * x - 1.

    # Downsampling block
    input_pyramid = None
    if self.progressive_input != 'none':
      input_pyramid = x
    for mod in self.init_modules:
      h = mod(x)
    for mod in self.all_modules:
      h = mod(h, temb,zemb)
    for i,mod in enumerate(self.fin_modules):
      if i==0:
        h = mod(h)
        h = self.act(h)
      else:
        h = mod(h)
#     hs = [modules[m_idx](x)]
#     m_idx += 1
#     for i_level in range(self.num_resolutions):
#       # Residual blocks for this resolution
#       for i_block in range(self.num_res_blocks):
#         h = modules[m_idx](hs[-1], temb, zemb)
#         # print(h.shape)
#         m_idx += 1
#         if h.shape[-1] in self.attn_resolutions:
#           h = modules[m_idx](h)
#           m_idx += 1

#         hs.append(h)

#       if i_level != self.num_resolutions - 1:
#         if self.resblock_type == 'ddpm':
#           h = modules[m_idx](hs[-1])
#           m_idx += 1
#         else:
#           h = modules[m_idx](hs[-1], temb, zemb)
#           m_idx += 1

#         if self.progressive_input == 'input_skip':
#           input_pyramid = self.pyramid_downsample(input_pyramid)
#           h = modules[m_idx](input_pyramid, h)
#           m_idx += 1

#         elif self.progressive_input == 'residual':
#           input_pyramid = modules[m_idx](input_pyramid)
#           m_idx += 1
#           if self.skip_rescale:
#             input_pyramid = (input_pyramid + h) / np.sqrt(2.)
#           else:
#             input_pyramid = input_pyramid + h
#           h = input_pyramid

#         hs.append(h)

#     h = hs[-1]
#     h = modules[m_idx](h, temb, zemb)
#     m_idx += 1
#     h = modules[m_idx](h)
#     m_idx += 1
#     h = modules[m_idx](h, temb, zemb)
#     m_idx += 1

#     pyramid = None

#     # Upsampling block
#     for i_level in reversed(range(self.num_resolutions)):
#       for i_block in range(self.num_res_blocks + 1):
#         h = modules[m_idx](torch.cat([h, hs.pop()], dim=1), temb, zemb)
#         m_idx += 1

#       if h.shape[-1] in self.attn_resolutions:
#         h = modules[m_idx](h)
#         m_idx += 1

#       if self.progressive != 'none':
#         if i_level == self.num_resolutions - 1:
#           if self.progressive == 'output_skip':
#             pyramid = self.act(modules[m_idx](h))
#             m_idx += 1
#             pyramid = modules[m_idx](pyramid)
#             m_idx += 1
#           elif self.progressive == 'residual':
#             pyramid = self.act(modules[m_idx](h))
#             m_idx += 1
#             pyramid = modules[m_idx](pyramid)
#             m_idx += 1
#           else:
#             raise ValueError(f'{self.progressive} is not a valid name.')
#         else:
#           if self.progressive == 'output_skip':
#             pyramid = self.pyramid_upsample(pyramid)
#             pyramid_h = self.act(modules[m_idx](h))
#             m_idx += 1
#             pyramid_h = modules[m_idx](pyramid_h)
#             m_idx += 1
#             pyramid = pyramid + pyramid_h
#           elif self.progressive == 'residual':
#             pyramid = modules[m_idx](pyramid)
#             m_idx += 1
#             if self.skip_rescale:
#               pyramid = (pyramid + h) / np.sqrt(2.)
#             else:
#               pyramid = pyramid + h
#             h = pyramid
#           else:
#             raise ValueError(f'{self.progressive} is not a valid name')

#       if i_level != 0:
#         if self.resblock_type == 'ddpm':
#           h = modules[m_idx](h)
#           m_idx += 1
#         else:
#           h = modules[m_idx](h, temb, zemb)
#           m_idx += 1

#     assert not hs

#     if self.progressive == 'output_skip':
#       h = pyramid
#     else:
#       h = self.act(modules[m_idx](h))
#       m_idx += 1
#       h = modules[m_idx](h)
#       m_idx += 1

#     assert m_idx == len(modules)
    
    if not self.not_use_tanh:

        return torch.tanh(h)
    else:
        return h